from tko.settings.settings import Settings
from tko.settings.repository import Repository
from uniplot import plot_to_string # type: ignore
from tko.util.text import Text
from tko.logger.log_sort import LogSort
from tko.logger.log_item_exec import LogItemExec
from tko.logger.delta import Delta

class TaskGraph:
    def __init__(self, settings: Settings, rep: Repository, task_key: str, width: int, height: int):
        self.settings = settings
        self.rep = rep
        self.task_key = task_key
        self.width = width
        self.height = height
        
        self.collected_rate: list[float] = []
        self.collected_elapsed: list[float] = []
        self.collected_lines: list[float] = []

        self.total_elapsed: float = 0.0
        self.max_lines: float = 0
        self.actual_rate: float = 0.0
        
        self.eixo: list[float] = []
        self.logger = rep.logger

    def __collect(self):
        key_dict: dict[str, LogSort] = self.logger.tasks.task_dict
        if not self.task_key in key_dict:
            return
        
        item_exec_list: list[tuple[Delta, LogItemExec]] = key_dict[self.task_key].exec_list
        collected_rate: list[float] = [0]
        collected_elapsed: list[float] = [0]
        collected_lines: list[float] = [0]
        last = 0
        eixo: list[float] = [0]
        count = 1
        last_size = 0
        for delta, item in item_exec_list:
            if item.rate == -1:
                collected_rate.append(last)
            else:
                last = item.rate
                collected_rate.append(last)
            collected_elapsed.append(delta.accumulated.total_seconds())  # Convert to minutes
            if item.size > 0:
                last_size = item.size
            collected_lines.append(last_size)
            eixo.append(count)
            count += 1
        self.total_elapsed = collected_elapsed[-1]
        self.max_lines = max(collected_lines)
        self.actual_rate = collected_rate[-1] if collected_rate else 0
        if collected_elapsed[-1] != 0:
            for i in range(len(collected_elapsed)):
                collected_elapsed[i] = collected_elapsed[i] / self.total_elapsed * 100
        if self.max_lines != 0:
            for i in range(len(collected_lines)):
                collected_lines[i] = (collected_lines[i] / self.max_lines) * 100
                if collected_lines[i] < 1:
                    collected_lines[i] = 0

        self.collected_rate = collected_rate
        self.collected_elapsed = collected_elapsed
        self.collected_lines = collected_lines
        self.eixo = eixo
        # self.eixo = list(range(len(collected)))

    def get_graph(self) -> list[Text]:
        self.__collect()
        if not self.eixo:
            return []
        title = Text.format(" {C}", f" @{self.task_key} ")
        title += Text.format(" {G}", f" Total {self.actual_rate:.0f}% ")
        time_h: int = int(self.total_elapsed) // 3600
        time_m: int = (int(self.total_elapsed) % 3600) // 60
        time = f"{time_h:02.0f}h {time_m:.0f}min" if time_h > 0 else f"{time_m:.0f}min"
        title += Text.format(" {B}", f" Tempo {time} ")
        title += Text.format(" {M}", f" Linhas {self.max_lines:.0f} ")
        # if len(self.collected_elapsed) > 1:
        result = plot_to_string(xs=[self.eixo, self.eixo, self.eixo], ys=[self.collected_elapsed, self.collected_lines, self.collected_rate], lines=[True, True, True], y_min=0, y_max=101, width=self.width, height=self.height, y_unit="%", x_unit="runs")

        if isinstance(result, str):
            result = result.splitlines()

        lines: list[Text] = []
        for line in result:
            lines.append(Text.decode_raw(line))
        title = title.center(self.width)
        lines.append(title)
        return lines
        # return []


from __future__ import annotations

from collections.abc import Callable

from .class_task import ClassTask
from .collected import Collected, Game
from tko.logger.task_resume import TaskResume, Resume
from tko.cmds.cmd_rep import CmdRep
from tko.game.task_info import TaskInfo
from tko.game.quest_grader import QuestGrader
import subprocess

import re
import os
import json
from .csv_tools import CSV

def update_reference_rep(target: str) -> None:
    """Update the reference repository for the given target."""
    class_task = ClassTask(target).load_from_file()
    reference = class_task.get_reference_rep()
    if reference is None:
        raise ValueError("Reference directory is not set in the class task.")
    if not os.path.exists(reference):
        raise FileNotFoundError(f"Reference directory not found: {reference}")
    cmd = f"tko rep {reference} update"
    subprocess.run(cmd, shell=True, check=True)


def col_one_to_str(entry: Game.Task | Game.Quest) -> str:
    if isinstance(entry, Game.Task):
        opening = "*" if entry.opt else "+"
        return f'{opening}:{entry.value}:{entry.key}'
    else:
        opening = "="
        return f'{opening}:{entry.value}:{entry.key}'

def calc_quest_percent(quest: list[Game.Task], task_dict: dict[str, TaskResume]) -> float:
    """Calculate the percent of a quest based on its tasks."""
    if not quest:
        return 0.0
    elements: list[QuestGrader.Elem] = []
    for task in quest:
        elements.append(QuestGrader.Elem(
            value=task.value,
            opt=task.opt,
            percent=task_dict.get(task.key, TaskResume(task.key)).resume.percent
        ))
    earned_xp, total_xp = QuestGrader.calc_xp_earned_total(elements)
    percent = QuestGrader.get_percent(earned_xp, total_xp)
    return percent

def task_resume_to_str(task_resume: TaskResume) -> str:
    return f'{round(task_resume.resume.percent):03d}'

def task_resume_to_full_str(task_resume: TaskResume) -> str:
    """Convert a TaskResume to a string with detailed information."""
    resume: Resume = task_resume.resume
    info: TaskInfo = task_resume.info
    return f'{info.rate:03d}% {info.flow}{info.edge}how {resume.minutes:03d}m {resume.executions:03d}/{resume.versions:03d}ex {round(resume.percent):03d}%'

def quest_resume_to_str(tasks: list[Game.Task], user_task_dict: dict[str, TaskResume]) -> str:
    percent = calc_quest_percent(tasks, user_task_dict)
    return f'{round(percent):03d}'

def clear_ansi_codes(text: str) -> str:
    """Remove ANSI escape codes from a string."""
    ansi_escape = re.compile(r'\x1B\[[0-?9;]*[ -/]*[@-~]')
    return ansi_escape.sub('', text)


class Grid:
    def __init__(self, user_info_map: dict[str, Collected]):
        self.col_one: list[Game.Cluster] = []
        self.inside: dict[str, dict[str, str]] = {}
        self.user_label_task: dict[str, dict[str, TaskResume]] = {}
        for folder, info in user_info_map.items():
            self.user_label_task[folder] = info.resume
        self.row_one: list[str] = sorted(self.user_label_task.keys())
        self.col_one_to_str: Callable[[Game.Task | Game.Quest], str] = col_one_to_str
        self.task_resume_to_str: Callable[[TaskResume], str] = task_resume_to_str
        self.quest_resume_to_str: Callable[[list[Game.Task], dict[str, TaskResume]], str] = quest_resume_to_str
        self.include_tasks: bool = True  # Include tasks in the grid
        self.include_quests: bool = True  # Include quests in the grid


    def set_col_one_to_str_function(self, func: Callable[[Game.Task | Game.Quest], str]) -> Grid:
        self.col_one_to_str = func
        return self

    def set_task_resume_to_str_function(self, func: Callable[[TaskResume], str]) -> Grid:
        self.task_resume_to_str = func
        return self

    def set_quest_resume_to_str_function(self, func: Callable[[list[Game.Task], dict[str, TaskResume]], str]) -> Grid:
        self.quest_resume_to_str = func
        return self

    def __calc_student_entries(self, user: str) -> dict[str, str]:
        user_task_dict: dict[str, TaskResume] = self.user_label_task[user]
        output: dict[str, str] = {}
        for ref_cluster in self.col_one:
            for ref_quest in ref_cluster.quests:
                output[ref_quest.key] = self.quest_resume_to_str(ref_quest.tasks, user_task_dict)
                for task in ref_quest.tasks:
                    if task.key in user_task_dict:
                        output[task.key] = self.task_resume_to_str(user_task_dict[task.key])
        return output

    def set_reference(self, reference: Collected):
        self.col_one: list[Game.Cluster] = reference.clusters
        return self
    
    def set_include_tasks(self, include_tasks: bool) -> Grid:
        self.include_tasks = include_tasks
        return self

    def set_include_quests(self, include_quests: bool) -> Grid:
        self.include_quests = include_quests
        return self

    def build_csv(self) -> list[list[str]]:
        for user in self.row_one:
            self.inside[user] = self.__calc_student_entries(user)
    
        data: list[list[str]] = [['SIMPLE']]
        col_one_keys: list[str] = ['SIMPLE']
        # coluna 1
        for cluster in self.col_one:
            for quest in cluster.quests:
                if self.include_quests:
                    data.append([self.col_one_to_str(quest)])
                    col_one_keys.append(quest.key)
                if self.include_tasks:
                    for task in quest.tasks:
                        data.append([self.col_one_to_str(task)])
                        col_one_keys.append(task.key)
        # linha 1
        for user in self.row_one:
            data[0].append(user)
        # print("\n".join(map(str, data)))

        # preencher o resto com campos vazios
        for i in range(1, len(data)):
            if len(data[i]) < len(data[0]):
                data[i].extend([''] * (len(data[0]) - len(data[i])))

        # preencher os dados
        for r in range(1, len(data)):
            for c in range(1, len(data[0])):
                user = data[0][c]
                quest_or_task_key = col_one_keys[r]
                if quest_or_task_key in self.inside[user]:
                    data[r][c] = self.inside[user][quest_or_task_key]
                else:
                    data[r][c] = '---'
        return data


class Join:
    @staticmethod
    def decode_students_collected_file(file: str) -> dict[str, Collected]:
        if not os.path.isfile(file):
            print(f"Collected data file not found: {file}")
            exit(1)
        with open(file, "r", encoding="utf-8") as f:
            data = json.load(f)
            user_info_map: dict[str, Collected] = {}
            for folder, text in data.items():
                collected_data = Collected().load_from_dict(json.loads(text))
                user_info_map[folder] = collected_data
        return user_info_map

    @staticmethod
    def join_student_daily_graph(user_info_map: dict[str, Collected]) -> str:
        lines: list[str] = []
        for folder, info in user_info_map.items():
            lines.append("─" * 120)
            lines.append(folder)
            lines.append(info.graph)
            lines.append("\n")
        return "\n".join(lines)

    @staticmethod
    def save_daily_graphs(class_task: ClassTask, user_info_map: dict[str, Collected]) -> None:
        graph_color, graph_mono = class_task.get_default_graph_joined()
        class_graph = Join.join_student_daily_graph(user_info_map)
        with open(graph_color, "w", encoding="utf-8") as f:
            f.write(class_graph)
        with open(graph_mono, "w", encoding="utf-8") as f:
            f.write(clear_ansi_codes(class_graph))

    @staticmethod
    def main(target: str) -> None:
        class_task = ClassTask(target).load_from_file()
        collected_path = class_task.get_default_collected()
        user_info_map = Join.decode_students_collected_file(collected_path)
        Join.save_daily_graphs(class_task, user_info_map)
        ref_folder = class_task.get_reference_rep()
        if ref_folder is not None:
            params = CmdRep.CollectParams()
            params.folder = ref_folder
            params.game = True
            params.json_output = True # dont echo
            reference: Collected = CmdRep.collect(params)
        else:
            reference: Collected = list(user_info_map.values())[0] # TODO

        # simple csv
        grid: Grid = Grid(user_info_map).set_reference(reference)
        path = class_task.get_default_simple_csv()
        CSV.save_csv(path, CSV.format_sheet(grid.build_csv()))

        # quest csv
        grid.set_include_tasks(False)
        path = class_task.get_default_quest_csv()
        CSV.save_csv(path, CSV.format_sheet(grid.build_csv()))

        # full csv
        grid.set_include_tasks(True).set_task_resume_to_str_function(task_resume_to_full_str)
        path = class_task.get_default_full_csv()
        CSV.save_csv(path, CSV.format_sheet(grid.build_csv()))



from .task_info import TaskInfo
import numpy as np

class TaskGrader:
    def __init__(self, task_info: TaskInfo):
        self.info = task_info

    @staticmethod
    def get_harmonic_mean(a: float, b: float, c: float) -> float:
        if a == 0 or b == 0 or c == 0:
            return 0.0
        return 3 / (1.0 / a + 1.0 / b + 1.0 / c)

    @staticmethod
    def get_square_mean(a: float, b: float, c: float) -> float:
        if a == 0 or b == 0 or c == 0:
            return 0.0
        return ((a * b * c) ** 0.5) / 10.0
    
    @staticmethod
    def get_cubic_square(a: float, b: float, c: float) -> float:
        if a == 0 or b == 0 or c == 0:
            return 0.0
        return ((a * b * c) ** (1/3))

    @staticmethod
    def get_product(a: float, b: float, c: float) -> float:
        if a == 0 or b == 0 or c == 0:
            return 0.0
        return (a * b * c) / 10000
    
    @staticmethod
    def get_sigmoid(a: float, b: float, c: float) -> float:
        x = a * b * c / 1000.0
        k = 0.011  # steepness of the curve
        x0 = 500.0
        v = 1 / (1 + np.exp(-k * (x - x0))) #numpy sigmoid function
        # v = 1 / (1 + (2.71828 ** (-k * (x - x0))))  # using e constant directly for compatibility
        return v * 100.0

    def get_edge_percent(self) -> int:
        edge = self.info.edge
        # deprecated
        if self.info.flow < 6 and edge == 5:
            edge = 4
        return edge * 20
    
    def get_flow_percent(self) -> int:
        if self.info.flow > 4:
            return 100
        if self.info.flow == 0:
            return 0
        return 80    

    def get_percent(self):
        # hardcoded gambi for compatibility
        rate = float(self.info.rate)
        flow = float(self.get_flow_percent())
        edge = float(self.get_edge_percent())

        # return Task.get_square_mean(flow, edge, rate)
        # return Task.get_harmonic_mean(flow, edge, rate)
        # return Task.get_cubic_square(flow, edge, rate)
        return TaskGrader.get_sigmoid(flow, edge, rate)


    def get_ratio(self) -> float:
        return self.get_percent() / 100.0



from __future__ import annotations
from tko.settings.repository import Repository
from tko.settings.settings import Settings
from tko.logger.logger import Logger
from tko.logger.task_resume import TaskResume
from tko.logger.log_sort import LogSort
from tko.play.week_graph import DailyGraph
from tko.mico.collected import Collected, Game
import yaml # type: ignore
import json
import os
import argparse
    

class CmdRep:
    class CollectParams:
        def __init__(self):
            self.folder: str = ""
            self.width: int = 10
            self.height: int = 10
            self.daily: bool = False
            self.resume: bool = False
            self.game: bool = False
            self.log: bool = False
            self.json_output: bool = False
            self.colored: int = 1

    @staticmethod
    def resume(rep: Repository) -> dict[str, TaskResume]:
        logger: Logger = rep.logger
        tasks: dict[str, LogSort] = logger.tasks.task_dict
        resume_dict: dict[str, TaskResume] = {}

        for key in tasks:
            log_sort = tasks[key]
            resume = TaskResume(key).from_log_sort(log_sort)
            resume_dict[key] = resume

        return resume_dict

    @staticmethod
    def graph(rep: Repository, width: int, height: int, colored: bool) -> str:
        dg = DailyGraph(rep.logger, width, height)
        image = dg.get_graph()
        if not colored:
            return "\n".join([x.get_str() for x in image])
        return "\n".join([str(x) for x in image])

    @staticmethod
    def load_game_as_cluster_list(rep: Repository) -> list[Game.Cluster]:
        game = rep.game
        if not game:
            return []
        output: list[Game.Cluster] = []
        for cluster_key in game.ordered_clusters:
            cluster = game.clusters[cluster_key]
            output_cluster = Game.Cluster(cluster.key)
            output.append(output_cluster)
            for quest in cluster.get_quests():
                output_quest = Game.Quest(quest.key, quest.value)
                output_cluster.quests.append(output_quest)
                for task in quest.get_tasks():
                    output_quest.tasks.append(Game.Task(key=task.key, value=task.xp, is_leet=task.is_leet(), opt=task.opt))
        return output

    @staticmethod
    def collect_main(args: argparse.Namespace):
        params = CmdRep.CollectParams()
        params.folder = args.folder
        params.width = args.width
        params.height = args.height
        params.daily = args.daily
        params.resume = args.resume
        params.game = args.game
        params.log = args.log
        params.json_output = args.json
        params.colored = args.color
        data = CmdRep.collect(params)

        if params.json_output:
            # print(yaml.dump(data))
            print(json.dumps(data.to_dict(), indent=4, ensure_ascii=False))

    @staticmethod
    def collect(param: CollectParams) -> Collected:
        rep = Repository(param.folder)
        if not rep.found():
            path = os.path.abspath(param.folder)
            print(f"Repository not found in {path}")
            return Collected()
        
        rep.load_config().load_game()
        data = Collected()

        if param.daily:
            graph = CmdRep.graph(rep, param.width, param.height, param.colored == 1)
            data.graph = graph
            if not param.json_output:
                print(graph)

        if param.resume:
            resume_data = CmdRep.resume(rep)
            data.resume = resume_data
            if not param.json_output:
                for key, value in resume_data.items():
                    print(f"{key}: {value.to_dict()}")

        if param.log:
            log_data = rep.logger.history.get_entries()
            data.log = [str(entry) for entry in log_data]
            if not param.json_output:
                for entry in log_data:
                    print(entry)

        if param.game:
            game_data = CmdRep.load_game_as_cluster_list(rep)
            data.clusters = game_data
            if not param.json_output:
                for cluster in game_data:
                    print(str(cluster))

        return data


    @staticmethod
    def update(args: argparse.Namespace):
        folder = args.folder
        if not os.path.isdir(folder):
            print(f"Folder {folder} does not exist.")
            return
        rep = Repository(folder)
        if not rep.found():
            print(f"Folder {folder} is not a valid tko repository.")
            return
        rep.load_config().load_game()
        print(f"Repositório cache atualizado.")

    @staticmethod
    def list(_args: argparse.Namespace):
        settings = Settings()
        print(f"SettingsFile\n- {settings.settings_file}")
        print(str(settings))

    @staticmethod
    def add(args: argparse.Namespace):
        settings = Settings().set_alias_remote(args.alias, args.value)
        settings.save_settings()

    @staticmethod
    def rm(args: argparse.Namespace):
        sp = Settings()
        if args.alias in sp.dict_alias_remote:
            sp.dict_alias_remote.pop(args.alias)
            sp.save_settings()
        else:
            print("Repository not found.")

    @staticmethod
    def reset(args: argparse.Namespace):
        _ = args
        sp = Settings().reset()
        print(sp.settings_file)
        sp.save_settings()

    # @staticmethod
    # def graph(args):
    #     rep = Repository(args.folder).load_config().load_game()
    #     rep.game.check_cycle()
    #     Graph(rep.game).generate()



#!/bin/env python

import os
import argparse
from tko.util.text import Text
from .class_task import ClassTask
from .student_repo import StudentRepo
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
import subprocess

class Pull:

    @staticmethod
    def run_git_command(directory: str, command_args: list[str]) -> tuple[str, str, int]:
        """
        Executa um comando Git em um diretório específico.
        Retorna stdout, stderr, e returncode.
        """
        # Certifique-se de que o comando Git é executado no diretório correto
        # cwd (current working directory) é essencial aqui
        try:
            process = subprocess.Popen(
                ['git'] + command_args,
                cwd=directory,  # Executa o comando no diretório do repositório
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True  # Para decodificar automaticamente para string
            )
            stdout, stderr = process.communicate()
            return stdout, stderr, process.returncode
        except FileNotFoundError:
            print(f"Erro: O comando 'git' não foi encontrado. Certifique-se de que o Git está instalado e no seu PATH.")
            return "", f"Erro: 'git' not found.", 127
        except Exception as e:
            print(f"Erro ao executar o comando 'git {' '.join(command_args)}': {e}")
            return "", f"Erro inesperado: {e}", 1


    @staticmethod
    def pull(directory: str) -> Text:
        """
        Pull a single repository in folter ignoring any local file or folder changes.
        If the repository needs a merge, ignore it and just pull the latest changes.
        Args:
            directory (str): The directory of the repository to pull.
        """
        if not os.path.isdir(directory):
            return Text()
        if not os.path.isdir(os.path.join(directory, ".git")):
            return Text()

        output: Text  = Text()

        output += Text("g") + directory + " "
        # 1. Reset Local Hard
        output += Text("y") + "Local "
        stdout, stderr, returncode = Pull.run_git_command(directory, ["reset", "--hard", "HEAD"])
        if returncode != 0:
            output += "\n" + f"Erro no reset --hard: {stderr}"


        # 2. Clean
        output += Text("y") + "Clean "
        stdout, stderr, returncode = Pull.run_git_command(directory, ["clean", "-fd"])
        if returncode != 0:
            output += "\n" + f"Erro no clean -fd: {stderr}"
        # else:
        #     print(f"  {stdout.strip()}")

        # 3. Fetch
        output += Text("y") + "Fetch "
        stdout, stderr, returncode = Pull.run_git_command(directory, ["fetch", "origin"])
        if returncode != 0:
            output += "\n" + f"Erro no fetch: {stderr}"
        # else:
        #     print(f"  {stdout.strip()}")

        # 4. Restore (git reset --hard origin/main)
        # Assumindo que o branch principal é 'main'. Mude para 'master' se for o caso.
        main_branch = "main" # Ou "master"
        output += Text("y") + "Restore "
        stdout, stderr, returncode = Pull.run_git_command(directory, ["reset", "--hard", f"origin/{main_branch}"])
        if returncode != 0:
            output += "\n" + f"Erro no restore (reset --hard origin/{main_branch}): {stderr}"
        else:
            output += "\n" + f"  {stdout.strip()}"

        return output

    @staticmethod
    def pull_all_parallel(repo_list: list[StudentRepo], max_workers: int = 50):
        """
        Pull all repositories in a list of student folders concurrently.
        Args:
            students_folders (list[str]): List of student folder paths.
            max_workers (int): Maximum number of threads to use.
        """
        print(f"\n{Text('y')}Iniciando pull de {len(repo_list)} repositórios com {max_workers} threads...{Text('e')}")
        start_time = time.time()

        # Use ThreadPoolExecutor para gerenciar as threads
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            # Cria um dicionário para mapear os objetos Future de volta para as pastas
            future_to_folder = {executor.submit(Pull.pull, repo.folder): repo.folder for repo in repo_list}

            # Itera sobre os resultados à medida que ficam prontos
            for future in as_completed(future_to_folder):
                student_folder = future_to_folder[future]
                try:
                    # Obtém o objeto GitOutput retornado pela função pull
                    output = future.result()
                    if output.get_str() != "":
                        print(output)
                except Exception as exc:
                    print(f"{Text('r')}A pasta '{student_folder}' gerou uma exceção não esperada: {exc}{Text('e')}")

        end_time = time.time()
        print(f"\n{Text('y')}Processamento concluído em {end_time - start_time:.2f} segundos.{Text('e')}")

    @staticmethod
    def main(class_task_path: str, n_threads: int = 5) -> None:
        class_task = ClassTask(class_task_path).load_from_file()
        students_repo_list = class_task.load_student_repo_list()
        Pull.pull_all_parallel(students_repo_list, n_threads)


def main():
    parser = argparse.ArgumentParser(description='Pull all git repositories in a directory')
    parser.add_argument('jsons', type=str, nargs="*", help='Directory to pull repositories from')
    parser.add_argument("-f", '--folders', type=str, nargs="*", help='Update git folder')
    parser.add_argument("-t", "--threads", type=int)

    args = parser.parse_args()
    n_threads = 10
    if args.threads:
        n_threads = args.threads

    if args.folders:
        Pull.pull_all_parallel(args.folders, n_threads)

    if args.jsons:
        for json in args.jsons:
            Pull.main(json, n_threads)


if __name__ == '__main__':
    main()



from __future__ import annotations
from tko.logger.old_log_loader import OldLogLoader
from tko.logger.log_item_base import LogItemBase
from tko.logger.log_item_self import LogItemSelf
from tko.logger.log_item_exec import LogItemExec
from tko.logger.log_item_move import LogItemMove
from tko.logger.delta import Delta
from tko.settings.rep_paths import RepPaths
import datetime as dt

from tko.util.decoder import Decoder
from typing import Callable
import os


class LogHistory:

    def __init__(self, rep_folder: str, listeners: list[Callable[[LogItemBase, bool], None]] | None = None):
        if listeners is None:
            listeners = []
        self.paths = RepPaths(rep_folder)
        self.log_folder: str = self.paths.get_log_folder()
        self.listeners: list[Callable[[LogItemBase, bool], None]] = listeners
        self.entries: list[LogItemBase] = []
        self.entries.extend(self.__load_old_log())
        new_log_entries = self.__load_daily_log_folder()
        # avoid duplicated entries
        item_index: dict[str, int] = {}
        for i, item in enumerate(self.entries):
            item_index[item.get_timestamp()] = i

        for item in new_log_entries:
            timestamp = item.get_timestamp()
            if timestamp in item_index:
                index = item_index[timestamp]
                self.entries[index] = item
            else:
                self.entries.append(item)
        
        for entry in self.entries:
            for listener in self.listeners:
                listener(entry, False)


    def __load_old_log(self) -> list[LogItemBase]:
        self.old_log_file = self.paths.get_old_history_file()
        loader = OldLogLoader(self.paths.get_rep_dir())
        return loader.base_list

    def get_entries(self) -> list[LogItemBase]:
        return self.entries

    def get_log_folder(self) -> str | None:
        return self.log_folder


    @staticmethod
    def log_file_for_day(folder: str, datetime: dt.datetime) -> str:
        date_str = datetime.strftime("%Y-%m-%d")
        if not os.path.exists(folder):
            os.makedirs(folder)
        return os.path.abspath(os.path.join(folder, f"{date_str}.log"))

    def append_new_action(self, item_base: LogItemBase) -> LogItemBase:
        now_str, now_dt = Delta.now()
        item_base.set_timestamp(now_str, now_dt)
        self.entries.append(item_base)
        for listener in self.listeners:
            listener(item_base, True)

        log_folder = self.get_log_folder()
        if log_folder is not None:
            log_file = LogHistory.log_file_for_day(log_folder, item_base.get_datetime())
            with open(log_file, 'a', encoding="utf-8", newline='') as file:
                file.write(f'{item_base.encode_line()}\n')
        return item_base

    def __load_daily_log_folder(self) -> list[LogItemBase]:
        log_folder = self.paths.get_log_folder()
        if not os.path.exists(log_folder):
            return []
        if not os.path.isdir(log_folder):
            raise ValueError(f"Log folder '{log_folder}' is not a directory.")
        files = os.listdir(log_folder)
        files_path = [os.path.join(log_folder, f) for f in files if f.endswith('.log')]
        if not files_path:
            return []
        # begin with the older file
        files_path.sort(key=lambda x: os.path.getmtime(x))
        entries: list[LogItemBase] = []
        for file in files_path:
            encoding = Decoder.get_encoding(file)
            with open(file, 'r', encoding=encoding) as f:
                for line in f:
                    item: LogItemBase | None = LogHistory.decode_line(line)
                    if item is not None:
                        entries.append(item)
        return entries
    
    @staticmethod
    def decode_line(line: str) -> LogItemBase | None:
        parts = line.strip().split(", ")
        if len(parts) < 2:
            return None
        item_type = LogItemBase.Type(parts[1])
        if item_type == LogItemBase.Type.MOVE:
            item = LogItemMove()
            if item.decode_line(parts):
                return item
        elif item_type == LogItemBase.Type.EXEC:
            item = LogItemExec()
            if item.decode_line(parts):
                return item
        elif item_type == LogItemBase.Type.SELF:
            item = LogItemSelf()
            if item.decode_line(parts):
                return item
        return None



# import yaml # type: ignore
# from tko.logger.log_item_base import LogItemBase
# from tko.logger.log_enum_item import LogEnumItem

# class DailyListener:
#     def __init__(self):
#         self.history: dict[str, dict[str, TaskBasic]] = {} # dict[day, dict[key, [key, cov, how]]]
#         self.actual: dict[str, TaskBasic] = {}
#         self.daily_file: str | None
        
#     def listener(self, action: LogItemBase, new_entry: bool = False):
#         if action.get_key() == "":
#             return
#         self.log_task(action.get_timestamp(), action.get_key(), action.get_cov(), action.get_app(), action.get_aut())

#     def set_daily_file(self, daily_file: str):
#         self.daily_file = daily_file
#         return self

#     def log_task(self, timestamp: str, key: str, coverage: int = -1, autonomy: int = -1, skill: int = -1):
#         if key not in self.actual:
#             self.actual[key] = TaskBasic(key)
#         self.actual[key].set_coverage(coverage).set_approach(autonomy).set_autonomy(skill)

#         day = timestamp.split(" ")[0]
#         if day not in self.history:
#             self.history[day] = {}
#         if key not in self.history[day]:
#             self.history[day][key] = self.actual[key]
#         else:
#             self.history[day][key].set_coverage(coverage).set_approach(autonomy).set_autonomy(skill)
#         # self.save_yaml()

#     def __str__(self) -> str:
#         output: dict[str, dict[str, str]] = {}
#         for day in sorted(self.history.keys()):
#             output[day] = {}
#             for key in self.history[day]:
#                 output[day][key] = str(self.history[day][key])
#         return yaml.dump(output)
    
#     def save_yaml(self):
#         if self.daily_file is None:
#             return
#         with open(self.daily_file, 'w') as file:
#             file.write(str(self))



# from typing import override
from tko.util.symbols import symbols
from tko.util.text import Text
from tko.game.tree_item import TreeItem
from tko.game.task_info import TaskInfo
from tko.game.task_grader import TaskGrader
import enum


class Task(TreeItem):
    str_index = "idx"

    class Types(enum.Enum):
        UNDEFINED = 0
        VISITABLE_URL = 1
        STATIC_FILE = 2 # static folder inside database
        REMOTE_FILE = 3 # url link do download file
        IMPORT_FILE = 4 # source folder outside database to import files

        # @override
        def __str__(self):
            return self.name

    def __init__(self):

        super().__init__()
        self.line_number = 0
        self.line = ""
        self.info = TaskInfo()
        self.grader = TaskGrader(self.info)
        self.main_idx: int = 0
        self.leet: bool = False # if the task is a leet task, rate must be marked by running test cases
        self.skills: dict[str, int] = {} # skills
        
        self.xp: int = 0
        self.opt: bool = False
        
        self.link = ""
        self.link_type: Task.Types = Task.Types.UNDEFINED

        self.quest_key = ""
        self.cluster_key = ""

        self.folder: str | None = None # the place where are the test case file
        self.__is_reachable = False
        self.default_min_value = 5 # default min grade to complete task

    def set_leet(self, value: bool):
        self.leet = value
        return self
    
    def is_leet(self) -> bool:
        return self.leet

    def set_reachable(self, reachable: bool):
        self.__is_reachable = reachable
        return self

    def is_reachable(self) -> bool:
        return self.__is_reachable

    def set_folder(self, folder: str):
        self.folder = folder
        return self
    
    def get_folder(self) -> str | None:
        return self.folder
    
    @staticmethod
    def decode_approach_autonomy(value: int) -> tuple[int, int]:
        opts = [(0, 0), (1, 1), (1, 2), (2, 2), (3, 2), (1, 3), (2, 3), (3, 3), (4, 3), (3, 4), (4, 4)]
        autonomy = opts[value][0]
        skill = opts[value][1]
        return autonomy, skill

    def decode_from_dict(self, value: str):
        value_list = value[1:-1]
        kv_dict: dict[str, str] = {}
        key_values = value_list.split(",")
        for kv in key_values:
            (k, val) = kv.split(":")
            kv_dict[k.strip()] = val.strip()
        
        self.info.load_from_kv(kv_dict)
        if self.str_index in kv_dict:
            self.main_idx = int(kv_dict[self.str_index])

        # deprecated
        for k, val in kv_dict.items():
            if k == "cov" or k == 'cove':
                self.info.rate = int(val)
            elif k == "aut" or k == "appr":
                self.info.flow = int(val)
            elif k == "hab" or k == "auto":
                self.info.edge = int(val)
            elif k == "desc":
                self.info.neat = int(val)
            elif k == "desire":
                self.info.cool = int(val)
            elif k == "effort":
                self.info.easy = int(val)

    def load_from_db(self, value: str):
        if value.startswith("{"):
            self.decode_from_dict(value)
        # deprecated
        elif ":" not in value:
            self.info.flow, self.info.edge = Task.decode_approach_autonomy(int(value))
        else:
            v = value.split(":")
            if len(v) == 3:
                self.info.flow, self.info.edge = Task.decode_approach_autonomy(int(v[0]))
                self.main_idx = int(v[1])
                self.info.rate = int(v[2])
            elif len(v) == 4:
                self.info.rate = (int(v[0]))
                self.info.flow = (int(v[1]))
                self.info.edge = (int(v[2]))
                self.main_idx = (int(v[3]))
            else:
                raise ValueError(f"Invalid task value format: {value}. Expected format is 'flow:edge:main_idx:rate' or '{self.str_index}:value'.")

    def save_to_db(self) -> str:
        kv_dict = self.info.get_filled_kv()
        if self.main_idx != 0:
            kv_dict[self.str_index] = str(self.main_idx)
        return "{" + ", ".join(f"{k}:{v}" for k, v in kv_dict.items()) + "}"

    def is_db_empty(self) -> bool:
        return len(self.info.get_filled_kv()) == 0

    def get_prog_color(self, min_value: None | int = None) -> str:
        if min_value is None:
            min_value = self.default_min_value
        prog = self.info.rate // 10
        if prog == 0:
            return "c"
        if prog < min_value:
            return "r"
        if prog < 10:
            return "y"
        if prog == 10:
            return "g"
        return "w"  

    def get_prog_symbol(self, min_value: None | int = None) -> Text:
        if min_value is None:
            min_value = self.default_min_value
        color = self.get_prog_color(min_value)
        prog = self.info.rate // 10
        if prog == 0:
            return Text().add("x")
        if prog < min_value:
            return Text().addf(color, str(prog))
        if prog < 10:
            return Text().addf(color, str(prog))
        if prog == 10:
            return Text().addf(color, symbols.check.text)
        return Text().add("0")

    def get_xp(self) -> int:
        if self.xp == 0:
            return 1
        return self.xp

    def get_percent(self) -> float:
        return self.grader.get_percent()

    def get_ratio(self) -> float:
        return self.grader.get_ratio()

    def is_complete(self):
        return self.grader.get_percent() >= 70

    def not_started(self):
        return self.grader.get_percent() == 0

    def in_progress(self):
        return 0 < self.grader.get_percent() < 100

    # @override
    def __str__(self):
        lnum = str(self.line_number).rjust(3)
        key = "" if self.key == self.title else self.key + " "
        return f"{lnum} key:{key} title:{self.title} skills:{self.skills} remote:{self.link} type:{self.link_type} folder:{self.folder}"

    def has_at_symbol(self):
        return any([s.startswith("@") for s in self.title.split(" ")])



from tko.game.cluster import Cluster
from tko.game.quest import Quest
from tko.game.task import Task
# from tko.game.graph import Graph

from tko.play.floating_grade import FloatingGrade
from tko.settings.settings import Settings

from tko.cmds.cmd_down import CmdDown
from tko.cmds.cmd_run import Run

from tko.util.param import Param

from tko.cmds.cmd_down import DownProblem

from tko.play.fmt import Fmt
from tko.play.floating import Floating
from tko.play.gui import Gui
from tko.play.opener import Opener

from tko.play.tasktree import TaskAction
from typing import Callable

from tko.logger.log_item_self import LogItemSelf
from tko.logger.log_item_move import LogItemMove

import os
import tempfile
import subprocess


class PlayActions:

    def __init__(self, gui: Gui):
        self.app = Settings().app
        self.settings = Settings()
        self.fman = gui.fman
        self.rep = gui.rep
        self.tree = gui.tree
        self.game = gui.game
        self.graph_opened: bool = False
        self.gui = gui

    def gen_graph_path(self) -> str:
        return os.path.join(self.rep.paths.get_rep_dir(), "graph.png")
        
    @staticmethod
    def open_link_without_stdout_stderr(link: str):
        outfile = tempfile.NamedTemporaryFile(delete=False)
        subprocess.Popen("python3 -m webbrowser -t {}".format(link), stdout=outfile, stderr=outfile, shell=True)

    @staticmethod
    def get_task_folder(task: Task) -> str:
        if task.folder is None:
            raise Exception("Folder não encontrado")
        return task.folder

    def open_code(self):
        obj = self.tree.get_selected_throw()
        if isinstance(obj, Task):
            task: Task = obj
            folder = self.rep.get_task_folder_for_label(task.key)
            if os.path.exists(folder):
                opener = Opener(self.settings).set_fman(self.fman)
                opener.set_target([folder]).set_language(self.rep.get_lang())
                opener.load_folders_and_open()
            else:
                self.fman.add_input(
                    Floating(self.settings, "v>")
                    .put_text("\nO arquivo de código não foi encontrado.\n")
                    .error()
                )
        else:
            self.fman.add_input(
                Floating(self.settings, "v>")
                .put_text("\nVocê só pode abrir o código")
                .put_text("de tarefas baixadas.\n")
                .error()
            )

    def open_link(self):
        obj = self.tree.get_selected_throw()
        if isinstance(obj, Task):
            task: Task = obj
            if task.link_type == Task.Types.VISITABLE_URL or task.link_type == Task.Types.REMOTE_FILE:
                try:
                    PlayActions.open_link_without_stdout_stderr(task.link)
                except Exception as _:
                    pass
            self.fman.add_input(
                Floating(self.settings, "v>")
                .set_header(" Abrindo link ")
                .put_text("\n " + task.link + " \n")
                .warning()
            )
        elif isinstance(obj, Quest):
            self.fman.add_input(
                Floating(self.settings, "v>")
                .put_text("\nEssa é uma missão.")
                .put_text("\nVocê só pode abrir o link")
                .put_text("de tarefas.\n")
                .error()
            )
        else:
            self.fman.add_input(
                Floating(self.settings, "v>")
                .put_text("\nEsse é um grupo.")
                .put_text("\nVocê só pode abrir o link")
                .put_text("de tarefas.\n")
                .error()
            )

    # def generate_graph(self):
    #     Graph(self.game).set_path(self.gen_graph_path()).set_opt(False).generate()
    #     path = self.gen_graph_path()
    #     if not self.graph_opened:
    #         opener = Opener(self.settings)
    #         opener.open_files([path])
    #         self.graph_opened = True

    def register_action(self, task: Task):
        self.rep.logger.store( LogItemSelf().set_key(task.key).set_info(task.info) )

    def evaluate(self):
        obj = self.tree.get_selected_throw()
        
        if isinstance(obj, Task):
            task: Task = obj
            self.fman.add_input(
                FloatingGrade(obj, self.settings, "").set_exit_fn(
                    lambda: self.register_action(task)
                )
            )
            return

    def down_remote_task(self):

        obj = self.tree.get_selected_throw()
        
        if isinstance(obj, Quest):
            self.fman.add_input(
                Floating(self.settings, "v>")
                .put_text("\nEssa é uma missão.")
                .put_text("\nVocê só pode baixar tarefas.\n")
                .error()
            )
            return
        if isinstance(obj, Cluster):
            self.fman.add_input(
                Floating(self.settings, "v>")
                .put_text("\nEsse é um grupo.")
                .put_text("\nVocê só pode baixar tarefas.\n")
                .error()
            )
        if not isinstance(obj, Task):
            return
        self.__down_remote_task(obj)
    
    def __down_remote_task(self, task: Task) -> None:
        if task.link_type != Task.Types.REMOTE_FILE and task.link_type != Task.Types.IMPORT_FILE:
            self.fman.add_input(
                Floating(self.settings, "v>").put_text("\nEssa não é uma tarefa de baixável.\n").error()
            )
            return
        lang = self.rep.get_lang() 
        down_frame = (
            Floating(self.settings, "v>").warning().set_text_ljust().set_header(" Baixando tarefa ")
        )
        # down_frame.put_text(f"\ntko down {self.rep.alias} {task.key} -l {lang}\n")
        self.fman.add_input(down_frame)

        def fnprint(text: str):
            down_frame.put_text(" " + text)
            down_frame.draw()
            Fmt.refresh()

        cmd_down = CmdDown(rep=self.rep, task_key=task.key, settings=self.settings)
        cmd_down.set_game(self.game)
        cmd_down.set_fnprint(fnprint)
        cmd_down.set_language(lang)
        result = cmd_down.execute()
        if result:
            self.rep.logger.store(
                LogItemMove().set_key(task.key).set_mode(LogItemMove.Mode.DOWN)
            )


    def select_task_action(self, task: Task) -> None:
        _, action = self.tree.get_task_action(task)
        if action == TaskAction.BAIXAR:
            self.__down_remote_task(task)
            return
        if action == TaskAction.VISITAR:
            self.open_link()
            return
        folder = task.get_folder()
        if not folder:
            raise Warning("Folder não encontrado")
        self.run_selected_task(task, folder)

    def select_task(self) -> Callable[[], None]:
        obj = self.tree.get_selected_throw()

        if isinstance(obj, Quest) or isinstance(obj, Cluster):
            self.tree.toggle(obj)
            return lambda: None

        if isinstance(obj, Task):
            task: Task = obj
            return lambda: self.select_task_action(task)
        return lambda: None

        
    def run_selected_task(self, task: Task, task_dir: str) -> None:
        folder = task_dir
        run = Run(settings=self.settings, target_list=[folder], param=Param.Basic())
        run.set_lang(self.rep.get_lang())
        opener = Opener(self.settings).set_language(self.rep.get_lang()).set_target([folder])
        run.set_opener(opener)
        run.set_run_without_ask(False)
        run.set_curses(True)
        run.set_task(self.rep, task)

        run.build_wdir()
        
        if not run.wdir.has_solver():
            DownProblem.create_default_draft(task_dir, self.rep.get_lang())
            msg = Floating(self.settings, "v>").warning()
            msg.put_text("\nNenhum arquivo de código na linguagem {} encontrado.".format(self.rep.get_lang()))
            msg.put_text("\nUm arquivo de rascunho foi criado em {}.".format(task_dir))
            self.fman.add_input(msg)
        run.execute()


from tko.game.quest import Quest, startswith
from tko.util.get_md_link import get_md_link


class QuestParser:
    quest: Quest

    def __init__(self):
        self.quest = Quest()
        self.line: str = ""
        self.line_num = 0
        self.filename: str = ""

    def finish_quest(self) -> Quest:

        if self.quest.key == "":
            self.quest.key = get_md_link(self.quest.title)
        return self.quest

    def match_full_pattern(self) -> bool:
        if not startswith(self.line, "### "):
            return False
        line = self.line[4:]

        pieces: list[str] = line.split("<!--")

        # html tags
        if len(pieces) > 1:
            middle_end: list[str] = pieces[1].split("-->")
            middle: str = middle_end[0]
            end: str = middle_end[1]
            line = pieces[0] + end # removendo raw text
            self.process_raw_tags(middle)

        self.quest.title = line
        if "[](" in line:
            pieces = line.split("[](")
            self.quest.title = pieces[0]

            del pieces[0]
            for p in pieces:
                key = p.split(")")[0]
                if key[0] == "#":
                    key = key[1:]
                self.quest.requires.append(key)

        return True

    def process_raw_tags(self, raw_tags: str):
        tags = [tag.strip() for tag in raw_tags.split(" ")]

        # skills
        skills = [t[1:] for t in tags if t.startswith("+")]
        if len(skills) > 0:
            self.quest.skills = {}
            for s in skills:
                try:
                    k, v = s.split(":")
                    self.quest.skills[k] = int(v)
                except ValueError:
                    self.quest.skills[s] = 1  # default value is 1 if not specified

        # languages
        languages = [t[2:] for t in tags if t.startswith("l:")]
        if len(languages) > 0:
            self.quest.languages = []
            for l in languages:
                self.quest.languages.append(l)
        self.quest.prog = "prog" in tags

        # quest percent
        qmin = [t[2:] for t in tags if t.startswith("q:")]

        # quest percent
        value = [t[2:] for t in tags if t.startswith("v:")]

        if len(value) > 0:
            try:
                self.quest.value = int(value[0])
            except ValueError:
                self.quest.value = 0

        if len(qmin) > 0:
            try:
                self.quest.qmin = int(qmin[0])
            except ValueError:
                self.quest.qmin = 50
            



    def parse_quest(self, filename: str, line: str, line_num: int) -> None | Quest:
        self.line = line
        self.line_num = line_num
        self.filename = filename

        self.quest.line = self.line
        self.quest.line_number = self.line_num
        self.quest.cluster_key = ""

        if self.match_full_pattern():
            return self.finish_quest()

        return None